'' This object provides declarations and routines to manipulate
'' named object code spaces in I2C EEPROMS.  It also handles a
'' persistent storage area kept in the last long of the boot
'' EEPROM area.  This is overwritten when a new program is
'' downloaded to the EEPROM by the IDE, but is relocated to the
'' next "page" of the boot EEPROM when this is available.  This
'' area keeps a map of which I/O pins are used for I2C busses.
'' If present, the system automatically probes those busses for
'' EEPROMs during initialization.  There is a copy of the bitmap
'' stored in every code space when it is either named or deleted,
'' but the only copy referenced is the one in the second or first
'' code space of the boot EEPROM.

'' Named object code spaces are currently always 32Kx8 with the
'' name of the program stored in the last 4 longs (16 bytes) of
'' the 32Kx8 "page".  The last long contains system information
'' including a checksum so the file name is actually 12 chars.

'' These routines should be able to support another storage scheme,
'' but others may require some form of space compaction which can
'' be slow with a limited speed serial bus.

CON
  busBitmap = $8000 - 2         ' Bitmap of pins with I2C busses
  checkMap  = $8000 - 3         ' Checksum for busBitmap
  checkSum  = $8000 - 4         ' Checksum for filename
  filename  = $8000 - 16        ' 12 character filename
  
  eepromDev = %1010_000_0       ' EEPROM device address

OBJ
  ldr   : "OS_loader"           ' System definitions
  data  : "OS_i2cObject"        ' EEPROM access via I2C
  vga   : "OS_vgaHiresText"     ' Hires VGA driver modified for OS use
  
PUB updateBitmapEEPROMs | i, t
'' This routine scans all the I2C busses identified in the stored
'' bitmap and fills in the EEPROM table based on what's present.
'' Any I2C busses that do not have EEPROMs are removed from the bitmap.
  result := getBusBitmap
  repeat i from 0 to 15
    if |<i & result <> 0
      if (t := probeEEPROMs(i << 1)) == 0
        result &= !|<i
      word[long[ldr#systemPtr]+ldr#sys_eeproms][i] := t
  updateBitmap(result)

PRI getBusBitmap | i, j, mem, chkM, c, bitmap
'' This routine checks the first 32K of the boot EEPROM for a valid
'' I2C pin bitmap.  If there's more than 32K, the second 32K "page"
'' is checked first.  If there's a valid bitmap, this is returned.
'' Otherwise, a default bitmap showing only the boot EEPROM is used.
'' If the boot EEPROM is not accessable, the bitmap will be zero.
'' The resulting bitmap is returned.
  data.init(ldr#i2cBootSDA,ldr#i2cBootSCL)
  if data.testBus == data#_None
    if data.devicePresent(eepromDev)
      mem := word[long[ldr#systemPtr]+ldr#sys_eeproms+ldr#i2cBootSCL]
      repeat i from 1 to 0
        if |<i & mem <> 0
          bitmap~
          chkM := data.readLocation(eepromDev,i<<15+checkMap,16,8)
          repeat j from 0 to 1
            c := data.readLocation(eepromDev,i<<15+busBitmap+j,16,8)
            bitmap := bitmap | (c << (8 * i))
            chkm -= c
          if chkM == $FF
            return bitmap
      return |<(ldr#i2cBootSCL >> 1)

PRI updateBitmap(bitmap) | i, j, mem, chkM, c
'' This routine updates the I2C bus bitmap stored in the boot
'' EEPROM.  If there's a second "page", the bitmap is recorded
'' there.  Otherwise, it's saved in the first (or only) 32K.
  data.init(ldr#i2cBootSDA,ldr#i2cBootSCL)
  mem := word[long[ldr#systemPtr]+ldr#sys_eeproms+ldr#i2cBootSCL]
  repeat i from 1 to 0
    if |<i & mem
      chkM := $FF
      repeat j from 0 to 1
        c := bitmap & $FF
        bitmap := bitmap >> 8
        chkM -= c
        data.writeLocation(eepromDev,i<<15+busBitmap+j,c,16,8)
        waitcnt(clkfreq / 200 + cnt)
      data.writeLocation(eepromDev,i<<15+checkMap,chkM,16,8)
      waitcnt(clkfreq / 200 + cnt)
      return

PUB markNewName(namePtr,addr) | pin, devA
'' This routine initializes the code space at the specified address
'' on the I2C bus with the supplied file name.  It does check that
'' the code space is either marked as deleted or is uninitialized.
'' The address format is pin pair in bits 21..19, page number in
'' bits 18..15, and address in 14..0.
  pin := addr >> 18 & $1E
  data.init(pin+1,pin)
  devA := %10100000 | ((addr >> 15) & %1110)
  addr &= $FFFF
  data.init(pin+1,pin)
  if checkSums(0,addr)
    writeFilename(namePtr,addr)
    return false
  abort -string("file not created, space not valid")

PUB markDeleted(namePtr,addr) | pin, devA
'' This routine marks as deleted the code space at the specified
'' address on the I2C bus.  It does check that the code space
'' has the specified name unless the pointer is zero.
'' The address format is pin pair in bits 21..19, page number in
'' bits 18..15, and address in 14..0.
  pin := addr >> 18 & $1E
  data.init(pin+1,pin)
  devA := %10100000 | ((addr >> 15) & %1110)
  addr &= $FFFF
  if checkSums(namePtr,addr)
    writeFilename(0,addr)
    return false
  abort -string("file not deleted, space not valid")

PUB searchFile(namePtr) | i, j, b, addr
'' This routine searches all available EEPROMs for an object code
'' space with the name given.  It returns the pin # in bits 22-19,
'' the 32K page number in bits 18-15, and the address in bits 14-0.
'' If the code space is not found, an abort is done with the
'' negative address of an error message returned.  If namePtr is
'' zero, this looks for an uninitialized or deleted code space.
  repeat i from 0 to 15          ' Pin pairs
    b := word[long[ldr#systemPtr]+ldr#sys_eeproms][i]
    repeat j from 0 to 15        ' 32Kx8 spaces
      if |<j & b <> 0            ' Is this one present?
        data.init(i<<1+1,i<<1)   ' Use pin pair
        addr := i << 19 | j << 15
        if checkSums(namePtr,addr)  ' Check name,chksums
          return addr            ' Match, return location
  abort -string("file not found")

PRI checkSums(namePtr, addr) | i, p, c, eq, chkS, chkM, devA
'' This routine checks the filename and bitmap checksums.
'' If those are correct, it compares the filename to the
'' one supplied (or to all blanks if none supplied) and
'' returns true if all match or if the checksums fail and
'' there was no name supplied (to catch the uninitialized
'' situation).
  devA := %10100000 | ((addr >> 15) & %1110)
  addr &= $FFFF
  chkM := data.readLocation(devA,addr+checkMap,16,8)
  repeat i from 0 to 1
    chkM += data.readLocation(devA,addr+busBitmap+i,16,8)
  if (p := namePtr) == 0
    p := string("            ")
  eq := true
  chkS := data.readLocation(devA,addr+checkSum,16,8)
  repeat i from 0 to 11
    c := data.readLocation(devA,addr+filename+i,16,8)
    chkS += c
    eq := eq and c == byte[p++]
  if chkS == $FF and chkM == $FF
    result := eq
  else
    result := namePtr == 0

PRI writeFilename(namePtr, addr) | i, p, c, chkS, chkM, m, devA, pin
  m := getBusBitmap             ' 
  pin := addr >> 18 & $1E
  data.init(pin+1,pin)
  devA := %10100000 | ((addr >> 15) & %1110)
  addr &= $FFFF
  chkM := $FF
  repeat i from 0 to 1
    c := m & $FF
    m := m >> 8
    chkM -= c
    data.writeLocation(devA,addr+busBitmap+i,c,16,8)
    waitcnt(clkfreq / 200 + cnt)
  data.writeLocation(devA,addr+checkMap,chkM,16,8)
  waitcnt(clkfreq / 200 + cnt)
  if (p := namePtr) == 0
    p := string("            ")
  chkS := $FF
  repeat i from 0 to 11
    c := byte[p++]
    chkS -= c
    data.writeLocation(devA,addr+filename+i,c,16,8)
    waitcnt(clkfreq / 200 + cnt)
  data.writeLocation(devA,addr+checkSum,chkS,16,8)
  waitcnt(clkfreq / 200 + cnt)

PUB probeEEPROMs(SCLpin) : bitmap | device, devA, addr, b0, b0a, b1, b1a, val, seed, t
'' This routine first attempts to see if an EEPROM is present.
'' A series of 32Kx8 address spaces could be provided by one
'' device or several.  This routine builds a bitmap showing which
'' have been provided if any and returns a value suitable for
'' placing in the system's EEPROM table.
  bitmap := 0
  seed := long[ldr#systemPtr] + ldr#sys_random
  data.init(SCLpin+1,SCLpin)
  if data.testBus == data#_none       ' Is there at least a pullup on SDA?
    repeat device from 0 to 7         ' If so, go through the EEPROM devices
      if data.devicePresent(eepromDev | (device << 1))
        bitmap |= |<(device * 2)      ' Add the ones that respond to the bitmap
    if bitmap <> 0                    ' If any responded, we have to see which
      repeat device from 0 to 7       ' devices are 32Kx8 and which are 64Kx8
        if bitmap & |<(device * 2) <> 0
          devA := eepromDev | (device << 1)
          addr := ?long[seed] & $7FFF ' Generate random 32K address
          b0 := data.readLocation(devA,addr,16,8)
          b1 := data.readLocation(devA,addr|$8000,16,8)
          repeat                      ' Generate random data not currently in
            val := ?long[seed] & $FF  ' the first 32Kx8 address space or in the
          until val <> b0 and val <> b1  ' the second address space (if present)
          data.writeLocation(devA,addr|$8000,val,16,8)  ' Write the data presumably
          waitcnt(clkfreq / 200 + cnt)  ' (5ms for write)
          b0a := data.readLocation(devA,addr,16,8)      ' to the second address space
          b1a := data.readLocation(devA,addr|$8000,16,8)  ' and see where it appears
          if b0 == b0a and b1a == val  ' If the change is in the second 32K only
            bitmap |= |<(device * 2 + 1)  ' then we really have memory there
            data.writeLocation(devA,addr|$8000,b1,16,8)  ' Restore the original data
          else
            data.writeLocation(devA,addr,b0,16,8)  ' If not there, restore first 32K
          waitcnt(clkfreq / 200 + cnt)  ' (5ms for write)

PUB dumpMemory(pin,page,addr,size) | i, c, p
'' This routine dumps a portion of the RAM or EEPROM to the display.
'' The format is 16 bytes wide with hexadecimal and printable characters.
'' If no EEPROM is present at that location, an error message is displayed.
  if pin <> -1
    data.init(pin+1,pin)
    if data.testBus <> data#_none
      vga.str(string("*** no I2C bus on those pins",vga#Cr))
      return
    if not data.devicePresent(page&%1110|eepromDev)
      vga.str(string("*** no EEPROM at that address",vga#Cr))
      return
  if addr + size > $7FFF               ' truncate to end of code space
    size := $8000 - addr
  p := addr & $7FF0
  repeat while p < (addr + size)
    vga.str(string("    "))
    if pin < (ldr#maxEEPROMs << 1)
      vga.decFixed(pin,2)
      vga.out(":")
      vga.decFixed(page,2)
      vga.out(" ")
    else
      vga.str(string("      "))
    vga.hex(addr+i,4)
    vga.str(string(vga#MoveX,64,"|"))
    repeat i from 0 to 15
      if p => addr and p < (addr + size)
        if pin < (ldr#maxEEPROMs << 1)
          c := data.readLocation(page&%1110|eepromDev,page&%0001<<15|p,16,8)
        else
          c := byte[p]
        vga.out2(vga#MoveX,3*i+15)
        vga.hex(c,2)
        vga.out2(vga#MoveX,i+65)
        if c < " " or c > "~"
          vga.out(".")
        else
          vga.out(c)
      p++
    vga.str(string(vga#MoveX,81,"|",vga#Cr))
  
PUB fileList | i, j, k, bitmap, devA, addr, chkM, chkS, a, b, c
'' This routine displays a list of all files currently known to the system.
  repeat i from 0 to 15          ' Pin pairs
    bitmap := word[long[ldr#systemPtr]+ldr#sys_eeproms][i]
    repeat j from 0 to 15        ' 32Kx8 spaces
      if (|<j & bitmap) <> 0     ' Is this one present?
        data.init(i<<1+1,i<<1)   ' Use pin pair
        devA := %10100000 | (j & %1110)  ' Device address
        addr := j & 1 << 15      ' 16 bit byte address
        chkM := data.readLocation(devA,addr+checkMap,16,8)
        repeat k from 0 to 1
          chkM += data.readLocation(devA,addr+busBitmap+k,16,8)
        chkS := data.readLocation(devA,addr+checkSum,16,8)
        repeat k from 0 to 11
          byte[@a][k] := data.readLocation(devA,addr+filename+k,16,8)
          chkS += byte[@a][k]
        vga.str(string("    "))
        vga.decFixed(i<<1,2)
        vga.out(":")
        vga.decFixed(j,2)
        if chkS == $FF and chkM == $FF
          vga.str(string(" <"))
          repeat k from 0 to 11
            vga.out(byte[@a][k])
          vga.str(string(">",vga#Cr))
        else
          vga.str(string(" Unused",vga#Cr))

PUB readEEPROM(addr,bytes) | pin, devA, i
'' This routine returns up to 4 bytes (LSB to MSB) from the specified location in an EEPROM.
'' The address format is pin pair in bits 21..19, page number in bits 18..15, address in 14..0.
  pin := addr >> 18 & $1E
  data.init(pin+1,pin)
  devA := %10100000 | ((addr >> 15) & %1110)
  addr &= $FFFF
  repeat i from 0 to (bytes <# 4) - 1
    result |= data.readLocation(devA,addr+i,16,8) << (i << 3)