'' This is the initialization (main) routine for the Propeller "Operating System"
'' This set of objects is intended to provide a basic set of I/O including VGA
'' text video, a PS/2 keyboard, and management of one or more I2C memory busses,
'' each containing up to 256K of EEPROM.  One is always the boot I2C bus on I/O
'' pins 28 and 29.  Others can be added on any pair of I/O pins.  The VGA pins
'' and keyboard pins are those used by the Propeller Demo Board, Rev C.  Included
'' is a SPIN loader that can load a normal SPIN program from one of the EEPROMs
'' known to the system, from any location in the EEPROM.  The size of the program
'' is stored with the program by the IDE, but this can be overridden.
'' The I/O drivers are intended to run continuously in 3 COGs (2 for video and 1
'' for the keyboard).  Their work areas (including buffers) are kept at the end
'' of RAM memory so they should not be overwritten when a new program is loaded.
'' The new program, if it uses these video text, keyboard, and loader objects,
'' should be able to seamlessly use the video and keyboard and transfer control
'' further to other programs and back to this main routine for further interaction.
'' The loader will stop any COGs not known to it prior to loading a new progam.
'' Programs can register COGs intended to "persist" between programs.  These
'' must be running assembly language and use a work area that won't be overlaid.
'' There are methods provided by the loader to do simple work area allocation.

'' All routines (except drivers where noted) are (c) 2006 by Michael Green.
'' Some I/O drivers are copyrighted by Parallax, some with modifications by
'' others.  See the notes at the beginning of each source file for specifics.

CON
  _clkmode    = xtal1 + pll16x
  _xinfreq    = 5_000_000

  vgaPinGroup = 16                             ' Base pin for VGA pin group
  ps2KeyData  = 26                             ' Data pin for PS/2 keyboard
  ps2KeyClock = 27                             ' Clock pin for PS/2 keyboard
  
  uniqueMark  = $965A3EC1                      ' Indicates that RAM was initialized

OBJ
  ldr     : "OS_loader"                        ' SPIN program loader and support routines
  vga     : "OS_vgaHiresText"                  ' Hires VGA driver modified for OS use
  vgaInit : "OS_vgaHiresTextInit"              '  Initialization and assembly portion
  key     : "OS_keyboard"                      ' PS/2 keyboard driver modified for OS use
  keyInit : "OS_keyboardInit"                  '  Initialization and assembly portion
  files   : "OS_fileSystem"                    ' File system support

CON
  _free = (ldr#endMemory - ldr#endFree + ldr#totalNeeded + 3) / 4  ' # longs needed
                                               
  #1, processProbe, processInfo,  processEdit  ' For command post-processing
      processHelp,  processRun,   processList
      processSave,  processLoad,  processDelete
      processDump

'' Internal SPIN information at fixed locations in RAM/EEPROM image

  pbase      = $0006                           ' must be $0010
  vbase      = $0008                           ' number of longs loaded times 4
  dbase      = $000A                           ' above where $FFF9FFFF's get placed
  pcurr      = $000C                           ' points to Spin code
  dcurr      = $000E                           ' points to local stack

VAR
  word dumpPin, dumpPage, dumpAddr, dumpSize
  long fileAddr, fileChange
  byte fileName[12]

PUB initialize | t, i, p
  if long[ldr#initMarker] <> uniqueMark                    ' Check for a valid initialization
    repeat i from 0 to 7
      if i <> cogid                                        ' Stop all cogs except this one
        cogstop(i)
    long[ldr#memPtr]    := ldr#endFree                     ' Start at end of HUB memory
    p := ldr.allocate(ldr#systemSize)                      ' Allocate system work area
    long[ldr#systemPtr] := p
    long[p+ldr#sys_cog] := 1 << cogid                      ' Mark current cog in use
    long[p+ldr#sys_random] := ldr#initMarker               ' Initialize random seed
    wordfill(p+ldr#sys_eeproms,0,ldr#maxEEPROMs)           ' Clear EEPROMs bitmap table
    long[ldr#videoPtr]  := ldr.allocate(ldr#videoSize)     ' Allocate work area for video
    long[ldr#screenPtr] := ldr.allocate(ldr#screenSize)    ' Allocate screen buffer
    long[ldr#colorPtr]  := ldr.allocate(ldr#colorSize)     ' Allocate screen color buffer
    long[ldr#cursorPtr] := ldr.allocate(ldr#cursorSize)    ' Allocate cursor control block
    if not vgaInit.start(vgaPinGroup)                      ' Now start up the VGA driver
      abort false
    vga.str(string("Propeller Operating System - Version 1.014.006",vga#Cr))
    long[ldr#keybdPtr]  := ldr.allocate(ldr#keybdSize)     ' Allocate work area for keyboard
    if not keyInit.start(ps2KeyData,ps2KeyClock)           ' Now start up the keyboard driver
      abort false
    long[ldr#initMarker] := uniqueMark                     ' Basic initialization complete
    waitcnt(clkfreq * 2 + cnt)                             ' Wait for keyboard to initialize
  if not key.present
    vga.str(string(vga#Cr,"--> keyboard not present, can't continue"))
    abort long[ldr#initMarker] := 0
  if files.updateBitmapEEPROMs & %0100000000000000 == 0    ' Check boot EEPROM usable/writable
    vga.str(string(vga#Cr,"--> boot eeprom not accessable, can't continue"))
    abort long[ldr#initMarker] := 0
  vga.out(vga#Cr)
  processCommands

PRI processCommands | t, y, i, j, z
  repeat
    y := vga.windowRow
    if (t := \enterDoCommand(y)) < 0
      vga.out4(vga#UseWin,0,vga#Cursor,0)
      vga.out4(vga#MoveXY,vga#cols,y,vga#Cr)
      vga.str(string("*** "))
      vga.str(-t)
      vga.out(vga#Cr)
    else
      vga.out4(vga#UseWin,0,vga#Cursor,0)
      vga.out4(vga#MoveXY,vga#cols,y,vga#Cr)
      case t
        processProbe:
          repeat i from 0 to 30 step 2
            z~
            y := word[long[ldr#systemPtr]+ldr#sys_eeproms+i]
            if y <> 0
              repeat j from 0 to 15
                if (|<j & y) <> 0
                  z += 32
              vga.str(string("    Pin "))
              vga.decFixed(i,2)
              vga.str(string("  "))
              vga.decFixed(z,3)
              vga.str(string("K",vga#Cr))
        processInfo:
           vga.str(string("    At "))
           vga.hex(fileAddr,6)
'          vga.str(string(", pbase="))
'          vga.hex(files.readEEPROM(fileAddr+pbase,2),4)
'          vga.str(string(", vbase="))
'          vga.hex(files.readEEPROM(fileAddr+vbase,2),4)
'          vga.str(string(", dbase="))
'          vga.hex(files.readEEPROM(fileAddr+dbase,2),4)
'          vga.str(string(", pcurr="))
'          vga.hex(files.readEEPROM(fileAddr+pcurr,2),4)
'          vga.str(string(", dcurr="))
'          vga.hex(files.readEEPROM(fileAddr+dcurr,2),4)
'          vga.out(vga#Cr)
        processEdit:
          editWindow
        processHelp:
          vga.str(string("    delete <file>",vga#Cr))
          vga.str(string("    dump <addr> <len>",vga#Cr))
          vga.str(string("    dump <pin> <page> <addr> <len>",vga#Cr))
          vga.str(string("    edit",vga#Cr))
          vga.str(string("    help",vga#Cr))
          vga.str(string("    info <file>",vga#Cr))
          vga.str(string("    list",vga#Cr))
          vga.str(string("    load <file>",vga#Cr))
          vga.str(string("    probe <pin>",vga#Cr))
          vga.str(string("    run <file>",vga#Cr))
          vga.str(string("    save <file>",vga#Cr))
        processList:
          files.fileList
        processDump:
          files.dumpMemory(dumpPin,dumpPage,dumpAddr,dumpSize)
        processSave:
          if fileChange
            vga.str(string("    File exists.  Overwrite it? (Y/n) "))
            repeat
              t := key.getkey
              if t == vga#Cr
                t := "y"
            until t == "Y" or t == "y" or t == "N" or t == "n"
            vga.out2(t,vga#Cr)
          else
            t := "y"
          if t == "Y" or t == "y"
            vga.str(string("    Save to "))
            vga.hex(fileAddr,6)
            vga.out(vga#Cr)
        processLoad:
          vga.str(string("    Load from "))
          vga.hex(fileAddr,6)
          vga.out(vga#Cr)
        processRun:
          vga.str(string("    Run from "))
          vga.hex(fileAddr,6)
          vga.out(vga#Cr)
        processDelete:
          vga.str(string("    Are you sure you want to delete it? (Y/n) "))
          repeat
            t := key.getkey
            if t == vga#Cr
              t := "y"
          until t == "Y" or t == "y" or t == "N" or t == "n"
          vga.out2(t,vga#Cr)
          if t == "Y" or t == "y"
            files.markDeleted(@fileName,fileAddr)
            vga.str(string("    Deleted at "))
            vga.hex(fileAddr,6)
            vga.out(vga#Cr)

PUB editWindow | t
  vga.out6(vga#SetWin,1,5,1,vga#cols-5,vga#rows-2)
  vga.out6(vga#SetWin,2,0,1,5,vga#rows-2)
  vga.out3(vga#UseWin,2,vga#ClrScr)
  repeat t from 1 to vga#rows - 2
    vga.out3(vga#MoveXY,0,t-1)
    vga.decFixed(t,4)
  vga.out5(vga#UseWin,1,vga#Cursor,%01111,vga#ClrScr)
  repeat
    t := vga.windowEditor
  until t == $CC
  vga.out5(vga#UseWin,0,vga#Cursor,%00000,vga#ClrScr)

PUB enterDoCommand(y) | t, x, a, b, c
'' Set up a window at the line specified and accept a command
'' line terminated by a Cr.  Scan for a command word and
'' command specific parameters, either words or numbers.
'' If the command is known and the parameters are valid,
'' execute the command.  Abort with a negative error string
'' address or return a positive command specific value.
  vga.out6(vga#SetWin,1,1,y,vga#cols-1,1)
  vga.out6(vga#UseWin,0,vga#MoveXY,0,y,">")
  vga.out5(vga#UseWin,1,vga#ClrScr,vga#Cursor,%01111)
  repeat
    t := vga.windowEditor
  until t == vga#Cr
  x~
  if (x := \getTextWord(x,@a,4,true," ")) < 0
    abort -string("no command given")
  case a
    $45564153:  ' save
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      fileChange := true
      if (fileAddr := \files.searchFile(@fileName)) < 0
        if (fileAddr := \files.searchFile(0)) < 0
          abort -string("can't create file")
        files.markNewName(@fileName,fileAddr)
        fileChange := false
      return processSave
    $424F5250:  ' probe
      if (x := \getNumber(x,@a)) < 0
        return processProbe
      if a < (ldr#maxEEPROMs << 1) and (a & 1 == 0)
        if (t := files.probeEEPROMs(a)) == 0
          abort -string("no eeprom(s) found on that pin pair")
        else
          word[long[ldr#systemPtr]+ldr#sys_eeproms+a] := t
          return processProbe
      else
        abort -string("pin number odd or does not exist")
    $204E5552:  ' run
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      if (fileAddr := \files.searchFile(@fileName)) < 0
        abort -string("file not found")
      return processRun
    $54494445:  ' edit
      return processEdit
    $4F464E49:  ' info
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      if (fileAddr := \files.searchFile(@fileName)) < 0
        abort -string("file not found")
      return processInfo
    $504C4548:  ' help
      return processHelp
    $5453494C:  ' list
      return processList
    $454C4544:  ' delete
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      if (fileAddr := \files.searchFile(@fileName)) < 0
        abort -string("file not found")
      return processDelete
    $44414F4C:  ' load
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      if (fileAddr := \files.searchFile(@fileName)) < 0
        abort -string("file not found")
      return processLoad
    $504D5544:  ' dump
      if (x := \getNumber(x,@a)) < 0
        abort -string("missing <pin> or <addr>")
      if (x := \getNumber(x,@b)) < 0
        abort -string("missing <page> or <len>")
      if (x := \getNumber(x,@c)) < 0
        if (a => 32768) or (b => 1024)
          abort -string("invalid number")
        dumpPin := dumpPage := ldr#maxEEPROMs << 1
        dumpAddr := a
        dumpSize := b
        return processDump
      else
        if (x := \getNumber(x,@t)) < 0
          abort -string("missing <len>")
        if (a => (ldr#maxEEPROMs << 1)) or (b => 8) or (c => 32768) or (t => 1024)
          abort -string("invalid number")
        dumpPin := a
        dumpPage := b
        dumpAddr := c
        dumpSize := t
        return processDump
    other:
      abort -string("unknown command")

PRI skipBlanks(x)
'' Skip blanks until at right border of window and return x
'' of first non-blank or abort with a negative error string
'' address.
  repeat
    if vga.windowAtEnd(x,0)
      abort -string("looking for non-blank at end of line")
  until vga.windowChar(x++,0) <> " "
  return x - 1

PRI getTextWord(x,textAddr,maxSize,uppercase,clearFirst) | c
'' Skip leading blanks, then copy up to maxSize non-blank
'' characters into the area at textAddr.  Return the
'' position of the next blank.  If uppercase is true,
'' store any lower case alphabetic characters as upper
'' case.  If clearFirst is not -1 (true), clear the
'' textAddr area to the value given.  If an error occurs,
'' abort with a negative error string address
  if clearFirst <> -1
    repeat c from 1 to maxSize
      byte[textAddr + c - 1] := clearFirst
  x := skipBlanks(x)
  repeat while (c := vga.windowChar(x++,0)) <> " "
    if maxSize > 0
      if uppercase and c => "a" and c =< "z"
        c := c & !%0100000
      byte[textAddr++] := c
      maxSize--
  return x - 1

PRI getNumber(x,numberAddr) | negative, c, t
'' Skip leading blanks, then convert a decimal number
'' with an optional sign ("+"/"-") into binary to be
'' stored in the long at numberAddr.  Return the
'' position of the next blank.  If an error occurs,
'' abort with a negative error string address.
'' Currently, there's no checking for 32 bit overflow.
  x := skipBlanks(x)
  c := vga.windowChar(x++,0)
  if (negative := c == "-") or c == "+"
    c := vga.windowChar(x++,0)
  if c == "$"
    c := vga.windowChar(x++,0)
    if (t := hexDigit(c)) < 0
      abort -string("invalid hex character")
    long[numberAddr] := t
    c := vga.windowChar(x++,0)
    repeat until (t := hexDigit(c)) < 0
      long[numberAddr] := long[numberAddr] << 4 | t
      c := vga.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if c == "%"
    c := vga.windowChar(x++,0)
    if not (c == "0" or c == "1")
      abort -string("invalid hex character")
    long[numberAddr] := c - "0"
    c := vga.windowChar(x++,0)
    repeat while c == "0" or c == "1"
      long[numberAddr] := long[numberAddr] << 1 | (c - "0")
      c := vga.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if not (c => "0" and c=< "9")
    abort -string("invalid numeric character")
  long[numberAddr] := c - "0"
  c := vga.windowChar(x++,0)
  repeat while c => "0" and c =< "9"
    long[numberAddr] := long[numberAddr] * 10 + c - "0"
    c := vga.windowChar(x++,0)
  if negative
    long[numberAddr] := -long[numberAddr]
  return x - 1

PRI hexDigit(c)
  if c => "0" and c =< "9"
    return c - "0"
  if c => "A" and c =< "F"
    return c - "A" + 10
  if c => "a" and c =< "f"
    return c - "a" + 10
  return -1