'' This object has an entry that takes an EEPROM address (truncated to the nearest 512 byte
'' boundary), the pin number of the SCL line for the EEPROM with the SDA line on the next
'' higher numbered pin, and a Boolean which is false if the program stored there is to be loaded
'' and executed, true if the current contents of the boot EEPROM are to be copied, first into RAM,
'' then into the specified area of the EEPROM.  The EEPROM address and pin numbers can specify
'' the boot EEPROM, but the copy option will just copy the boot area to itself through RAM unless
'' the EEPROM is larger than 32K.  If there is no EEPROM connected to the pins specified or the
'' address is out of range, this object may eventually just stop its cog.  Some EEPROM devices
'' will ignore extra address bits and the effect on the boot process is unpredictable.

'' This object also provides constant declarations for sizes of various system tables and
'' buffers and offsets to their pointers in "high memory".  The offsets within the system
'' work are are declared here as well.  Some simple methods are provided for memory allocation
'' and for registering and unregistering COGs as being persistant.  Any COGs not registered
'' will be stopped when a program is read into RAM.

CON
  maxEEPROMs  = 16                             ' Maximum number of possible I2C busses
  i2cBootSCL  = 28                             ' I2C Clock pin for boot EEPROM
  i2cBootSDA  = 29                             ' I2C Data pin for boot EEPROM

'' System work area definitions

  sys_cog     = 0                              ' Contains bitmap for cogs used by system
  sys_random  = 4                              ' Random seed for "?" operator
  sys_eeproms = 8                              ' Table of EEPROMs, one for each I/O pair
                                               ' Each word entry contains a bitmap of the
                                               ' 32Kx8 address blocks that are present on
                                               ' the I2C bus on the corresponding pair of
                                               ' I/O pins (if any).  Size = maxEEPROMs * 2

'' System work areas, buffer, and table sizes

  systemSize  = 40                             ' # of bytes for system information
  videoSize   = 24 + (16 * 8)                  ' # of bytes for video
  screenSize  = 128 * 64                       ' # of bytes for screen buffer
  colorSize   = 64 * 2                         ' # of bytes for color buffer
  cursorSize  = 8                              ' # of bytes for screen cursor info
  keybdSize   = 80                             ' # of bytes needed for keyboard
  videoNeeded = videoSize + screenSize + colorSize + cursorSize
  totalNeeded = systemSize + videoNeeded + keybdSize

'' The following are fixed locations assigned at the top of RAM memory to avoid
'' having them overwritten when a new program is loaded.  Most of these are pointers
'' to work areas allocated "below" them.  The last pointer is reserved for non-system
'' programs

  endMemory   = $8000                          ' Address past the end of RAM
  initMarker  = endMemory -  4                 ' Unique marker for initialized data
  memPtr      = endMemory -  8                 ' Address of last allocated long
  systemPtr   = endMemory - 12                 ' Address of system information
  videoPtr    = endMemory - 16                 ' Address of video work area
  screenPtr   = endMemory - 20                 ' Address of screen buffer
  colorPtr    = endMemory - 24                 ' Address of color buffer
  cursorPtr   = endMemory - 28                 ' Address of screen cursor info
  keybdPtr    = endMemory - 32                 ' Address of keyboard work area
  userPtr     = endMemory - 36                 ' Address of work area for "user"
  endFree     = userPtr                        ' Address past initial free memory

'' Local constants

  stopWhenDone  = $4000_0000
  signalDone    = $8000_0000
  bootWhenDone  = $C000_0000
  writeToEEPROM = $2000_0000
  bootEEPROMpin = 28

PUB loadExecuteBoot                                    ' Load/execute 1st 32K of boot EEPROM
  boot(0,bootEEPROMpin,false)

PUB allocate(size)                                     ' Allocate block of size bytes
  return long[memPtr] := (long[memPtr] - size) & !3    ' Always a long address

PUB deallocate                                         ' User work area no longer in use
  long[memPtr] := long[keybdPtr]

PUB registerCOG(bitmap)                                ' Register "persistant" COGs
  long[long[systemPtr]+sys_cog] |= bitmap

PUB unregisterCOG(bitmap)                              ' Mark COGs no longer "persistant"
  long[long[systemPtr]+sys_cog] &= !bitmap

PUB boot(address, SCLpin, copyEEPROM) | params
'' Starting at address (truncated to 1K boundary) in EEPROM attached to SCLpin
'' and SDApin one higher, either read a SPIN program into RAM and execute or
'' copy the contents of the boot EEPROM to the specified EEPROM space.  This is
'' intended to support loading a program from the IDE to the first 32K of the
'' boot EEPROM and loading a short program using the IDE into RAM to copy the
'' first program to some other place.
  params := (address & $7FC00 >> 3) | (SCLpin & $1E << 2) | (copyEEPROM & %100)
  coginit(cogid,@loadHubMemory,params)

DAT
                        org     0
loadHubMemory           mov     i2cTemp,0               ' Compute clock ticks for bit timing
                        mov     i2cShort,#0
:computeShort           cmpsub  i2cTemp,i2cTimeUnit wr,wc
                if_c    add     i2cShort,#1
                if_c    jmp     #:computeShort          ' First i2cShort := clkfreq / 4_000_000
                        mov     i2cClkLow,i2cShort
                        shl     i2cClkLow,#1            ' Then i2cClkLow := i2cShort * 2
                        mov     i2cClkHigh,i2cClkLow    '  and i2cClkHigh := i2cClkLow
                        mov     i2cTemp,PAR
                        test    i2cTemp,#%100    wz     ' Copy to to EEPROM or boot from EEPROM?
                if_z    mov     i2cSCL,i2cBootSCLm
                if_z    mov     i2cAddr,#0
                if_nz   call    #processPAR             ' Get EEPROM address and SCL/SDA masks
                        mov     i2cCount,#$10
                        call    #readLocation           ' Get preamble first (with length)
                if_nc   jmp     #:loadHubError
                        rdword  i2cCount,#$08           ' Get byte count from the loaded program
                        sub     i2cCount,#$10
                        call    #readLocation           ' Now get the rest of the program
                if_nc   jmp     #:loadHubError
                        mov     i2cTemp,PAR             ' Copy program to EEPROM?
                        test    i2cTemp,#%100    wz
                if_z    jmp     #:bootMemory
                        call    #processPAR             ' Get EEPROM address and SCL/SDA bit masks
                        rdword  i2cCount,#$08           ' Get byte count from the loaded program
                        call    #writeLocation          '  and write number of bytes necessary
                if_nc   jmp     #:loadHubError
:bootMemory             rdbyte  bootTemp,#$0004         ' If xtal/pll enabled, start up now
                        and     bootTemp,#$F8           '  while remaining in current mode
                        clkset  bootTemp
:delay                  djnz    time_xtal,#:delay       ' Allow 20ms@20MHz for xtal/pll to settle
                        rdbyte  bootTemp,#$0004         ' Switch to selected clock
                        clkset  bootTemp
                        coginit interpreter             ' Reboot this cog with interpreter

:loadHubError           cogid   i2cTemp                 ' If an error occurs, stop the cog
                        cogstop i2cTemp

processPAR              mov     i2cAddr,PAR
                        andn    i2cAddr,#%1111111       ' Isolate and position EEPROM address bits
                        shl     i2cAddr,#3
                        mov     i2cTemp,PAR             ' Isolate the SCL pin number
                        shr     i2cTemp,#2
                        and     i2cTemp,#%11110         ' Generate bit mask for SCL
                        mov     i2cSCL,#1
                        shl     i2cSCL,i2cTemp
                        mov     i2cSDA,i2cSCL           ' Generate bit mask for SDA from SCL
                        shl     i2cSDA,#1
                        mov     i2cBufAdr,#0            ' Start at RAM location 0
processPAR_ret          ret

i2cTimeUnit             long    4_000_000               ' 250ns - Setup and hold times
i2cBootSCLm             long    $10000000               ' Boot EEPROM clock line = pin 28
time_xtal               long    20 * 20000 / 4 / 1      ' 20ms (@20MHz, 1 inst/loop)
interpreter             long    $0001 << 18 + $3C01 << 4 + %0000
bootTemp                long    0

'' Mid level I2C routines.

'' Read a sequence of bytes from an attached I2C EEPROM into a buffer in hub memory.  The
'' starting I2C address is provided as well as a byte count and starting cog address.  Bytes
'' are packed LSB..MSB with any left over bytes in LSB..MSB order in the lower portion of
'' the last long.  The caller sets up the masks for the SCL and SDA pins and the timing
'' information for the bus speed required.  EEPROM addresses are assumed for Atmel or
'' Microchip or equivalent devices in the 32Kx8 to 512Kx8 range with the high order 3 bits
'' in the device address byte and the low order 16 bits in the device register bytes (MSB,LSB).
'' Exit with (c) clear if a NAK was received when not expected.  The write routine works
'' similarly except that it allows for paged writes.  Since all the devices in the range
'' supported (32Kx8 to 512Kx8) have pages of at least 64 bytes, this is the size supported.
'' When the address crosses a page boundary, a Stop Sequence is forced, then the device is
'' readdressed about 5ms later to allow for the write cycle to complete.  For sequential reads,
'' the smallest device is assumed to be 32Kx8 and, when reads cross a 32K boundary, a Stop
'' Sequence is forced and the device is readdressed.

readLocation            call    #i2cSendAddr            ' Address the device
                if_nc   jmp     #readDoStop             '  as if for a write
                        or      i2cData,#1
                        call    #i2cStart               ' Reselect the device
                        call    #i2cWrite               '  for read cycles
                if_nc   jmp     #readDoStop
                        mov     i2cData,#0
readNext                mov     i2cMask,#%10000000
                        mov     i2cTemp,i2cAddr         ' Is this the last byte in a
                        and     i2cTemp,i2c32kPage      '  32K page?  If so, force a
                        cmp     i2cTemp,i2c32kPage wz   '   NAK with this byte
                        muxz    i2cTemp,31   nr,wc
                if_nc   cmp     i2cCount,#1     wz      ' Is this the last byte to read?
                if_nc   muxz    i2cTemp,#1   nr,wc      ' If so, clear (c) so will NAK
                        call    #i2cRead
                        wrbyte  i2cData,i2cBufAdr       ' Write the byte to hub memory
                        add     i2cBufAdr,#1
                        add     i2cAddr,#1              ' Increment the addresses and
                        djnz    i2cCount,#readCheck     '  decrement the byte count
readDoStop              call    #i2cStop                ' Do a Stop Sequence if done
readLocation_ret        ret

readCheck               test    i2cAddr,i2c32kPage      ' Is this the last byte in a
                if_z    jmp     #readNext               '  32K page?  If so, force a
                        call    #i2cStop                '   Stop Sequence, then start
                        jmp     #readLocation           '    over in another device

writeLocation           call    #i2cSendAddr            ' Address the device
                if_nc   jmp     #writeDoStop
                        mov     i2cMask,#%10000000
writeNext               rdbyte  i2cData,i2cBufAdr       ' Read the byte from hub memory
                        call    #i2cWrite               '  and transfer to the EEPROM
                if_nc   jmp     #writeDoStop
                        add     i2cBufAdr,#1            ' Increment the addresses and
                        add     i2cAddr,#1              '  decrement the byte count
                        djnz    i2cCount,writeCheck
                        call    #i2cStop                ' Do a Stop Sequence if done
writeDoStop             mov     i2cdata,#%10100000
                        call    #i2cStart               ' Now do acknowledge polling ...
                        call    #i2cWrite               ' Try to setup a write.  If not
                if_nc   jmp     #writeDoStop            '   acknowledged, keep trying.
                        call    #i2cStop
writeLocation_ret       ret

writeCheck              test    i2cAddr,#%111111  wz    ' Did we roll over the end of a page?
                if_z    jmp     #writeNext              ' If not, continue with the next byte
                        call    #i2cStop                ' If so, generate a Stop Sequence
:writeWait              mov     i2cdata,#%10100000
                        call    #i2cStart               ' Now do acknowledge polling ...
                        call    #i2cWrite               ' Try to setup a write.  If not
                if_nc   jmp     #:writeWait             '   acknowledged, keep trying.
                        call    #i2cStop
                        jmp     #writeLocation          ' Now start over on the new page

i2cSendAddr             mov     i2cTemp,i2cAddr
                        and     i2cTemp,#%11111111
                        shl     i2cTemp,#16             ' Bits 7-0 of address sent as last byte
                        mov     i2cData,i2cAddr
                        and     i2cData,#%11111111
                        shl     i2cData,#16             ' Bits 15-8 of address sent as first byte
                        or      i2cData,i2cTemp
                        mov     i2cTemp,i2cAddr
                        shr     i2cTemp,#16             ' Bits 18-16 of address in device address
                        shl     i2cTemp,#1
                        or      i2cTemp,#%10100000
                        or      i2cData,i2cTemp
                        call    #i2cStart
                        call    #i2cWrite
                if_nc   jmp     #i2cSendAddr_ret
                        call    #i2cWrite
                if_nc   jmp     #i2cSendAddr_ret
                        call    #i2cWrite
i2cSendAddr_ret         ret

i2c32kPage              long    %0111_1111_1111_1111    ' 32K address boundary
i2cAddr                 long    0                       ' Starting address in EEPROM
i2cCount                long    0                       ' Number of bytes to read / write
i2cBufAdr               long    0                       ' Buffer address in cog memory
i2cTemp                 long    0

'' Low level I2C routines.  These are designed to work either with a standard I2C bus
'' (with pullups on both SCL and SDA) or the Propellor Demo Board (with a pullup only
'' on SDA).  Timing can be set by the caller to 100KHz, 400KHz, or 1MHz.

'' Do I2C Start Sequence.  This assumes that both SCL and SDA are floating (inputs).
'' The start sequence is where SDA goes from HIGH to LOW while SCL is HIGH.

i2cStart                mov     i2cTime,i2cShort        ' Prepare for initial timing
                        andn    outa,i2cSDA             ' Prepare to drive SDA low
                        or      outa,i2cSCL
                        or      dira,i2cSCL             ' Active drive SCL high
                        add     i2cTime,cnt
                        waitcnt i2cTime,i2cShort        ' Wait for setup time and      
                        or      dira,i2cSDA             '  active drive SDA low
                        waitcnt i2cTime,i2cClkLow       ' Wait for hold time and return
                        andn    outa,i2cSCL             '  with SCL low and SDA low
i2cStart_ret            ret                             

'' Do I2C Stop Sequence.  This assumes that SCL is low and SDA is indeterminant.
'' The stop sequence is where SDA goes from LOW to HIGH while SCL is HIGH.
'' i2cStart must have been called prior to calling this routine for initialization.

i2cStop                 or      dira,i2cSDA             ' Active drive SDA low
                        or      outa,i2cSCL             ' Active drive SCL high
                        or      dira,i2cSCL
                        waitcnt i2cTime,i2cShort        ' Wait for setup time and
                        andn    dira,i2cSDA             '  let SDA float to high
                        waitcnt i2cTime,i2cClkLow       ' Wait for hold time and
                        andn    dira,i2cSCL             '  let SCL float to high
                        waitcnt i2cTime,i2cClkLow       ' Wait for minimum bus free
                        or      i2cTemp,#1    nr,wc     '  time and force (c) true
i2cStop_ret             ret

'' Write I2C data.  This assumes that i2cStart has been called and that SCL is low,
'' SDA is indeterminant, and i2cTime is the time to the next SCL low->high transition.
'' The (c) flag will be set from the ACK/NAK bit on exit.

i2cWrite                mov     i2cMask,#%10000000
:i2cWriteBit            waitcnt i2cTime,i2cClkHigh      ' Wait for minimum SCL low
                        or      outa,i2cSCL             '  and drive SCL high
                        test    i2cData,i2cMask wz
              if_z      or      dira,i2cSDA             ' Copy data bit to SDA
              if_nz     andn    dira,i2cSDA
                        waitcnt i2cTime,i2cClkLow       ' Wait for minimum SCL high
                        andn    outa,i2cSCL             '  and drive SCL low
                        shr     i2cMask,#1      wz      ' Go do next bit if not done
              if_nz     jmp     #:i2cWriteBit
                        andn    dira,i2cSDA             ' Switch SDA to input and
                        waitcnt i2cTime,i2cClkHigh      '  wait for minimum SCL low
                        or      dira,i2cSCL             ' Take SCL high and wait for
                        waitcnt i2cTime,i2cClkLow       '  minimum high time, then
                        test    i2cSDA,ina      wc      '   sample SDA (ACK/NAK)
                        andn    outa,i2cSCL             '    and take SCL low
                        shr     i2cData,#8              ' Prepare for multibyte write
i2cWrite_ret            ret

'' Read I2C data.  This assumes that i2cStart has been called and that SCL is low,
'' SDA is indeterminant, and i2cTime is the time to the next SCL low->high transition.
'' ACK/NAK will be copied from the (c) flag on entry.

i2cRead                 mov     i2cMask,#%10000000
                        shl     i2cData,#8              ' Prepare for multibyte read
:i2cReadBit             waitcnt i2cTime,i2cClkHigh      ' Wait for minimum SCL low,
                        or      outa,i2cSCL             '  drive SCL high, and make
                        andn    dira,i2cSDA             '   sure SDA is set to input
                        waitcnt i2cTime,i2cClkLow       ' Wait for minimum SCL high,
                        test    i2cSDA,ina      wz      '  sample SDA just before
                        andn    outa,i2cSCL             '   taking SCL low
              if_nz     or      i2cData,i2cMask         ' Accumulate data bits
              if_z      andn    i2cData,i2cMask
                        shr     i2cMask,#1      wz      ' Shift the bit mask and
              if_nz     jmp     #:i2cReadBit            '  continue until done
                        waitcnt i2cTime,i2cClkHigh      ' Wait for end of SCL low
                        or      outa,i2cSCL             '  and take SCL high
              if_nc     andn    dira,i2cSDA             ' Copy the ACK/NAK bit to SDA
              if_c      or      dira,i2cSDA
                        waitcnt i2cTime,i2cClkLow       ' Wait for minimum SCL high
                        andn    outa,i2cSCL             '  and take SCL low
i2cRead_ret             ret

'' Local variables for low level I2C routines

i2cSCL                  long    0                       ' Bit mask for SCL
i2cSDA                  long    0                       ' Bit mask for SDA
i2cTime                 long    0                       ' Used for timekeeping
i2cShort                long    0                       ' Clock ticks for setup/hold times
i2cClkHigh              long    0                       ' Clock ticks for minimum clock high
i2cClkLow               long    0                       ' Clock ticks for minimum clock low
i2cData                 long    0                       ' Data to be transmitted / received
i2cMask                 long    0                       ' Bit mask for bit to be tx / rx